#!/bin/sh -x
#
# Script to build mpgedit RPMS from a source code bundle. 
# How to use
# 1) First, run the production build script make_dist.sh to
#    successful completion
# 2) Copy or symlink the source package into the "src" directory
#    of the production build tree.  The source package is named something
#    like mpgedit_0-71dev_src.exe.
# 3) Become root.
# 4) Run ./dist_rpm.sh rev_num
#    Note: Unless you modify all of the *.spec files rev numbers to match
#    the rev you are building, just use the version that matches the spec
#    rev number you have.
# 5) Sit back and watch the fun. 
# 6) Do this on jake for the RedHat 6 RPM build first.
# 7) Repeat this on batman for the RedHat 7 RPM build.
#
RPMBASE="/usr/src/redhat"
SRC="mpgedit_0-72beta_src"
MAJ="0.72"
REV="1"
RPM="rpmbuild"
host=`hostname | sed 's/\..*//'`
if [ "$host" = jake ]; then
  RPM="rpm"
  TARGET="6"
else
  TARGET="7"
fi


# Checkpoint release number is mandatory
#
if [ -z "$1" ]; then
  echo "ERROR: Must specify checkpoint release version number"
  exit 1
fi
RELEASE="$1"
shift

# Revision number is optional.  Default value is 1.
#
if [ -n "$1" ]; then
  REV="$1"
  shift
fi

# mpgedit-0.71.49-1.i386.spec  mpgedit-man-0.71.49-noarch.spec \
# xmpgedit-0.71.49-1.i386.spec
MPGEDITMAN="mpgedit-man-$MAJ.${RELEASE}-noarch.spec"
MPGEDIT="mpgedit-$MAJ.${RELEASE}-${REV}.i386.spec"
XMPGEDIT="xmpgedit-$MAJ.${RELEASE}-${REV}.i386.spec"

#mpgedit-man-0.71-49.i386.rpm
MPGEDITMANRPM="mpgedit-man-$MAJ-${RELEASE}.i386.rpm"
MPGEDITRPM="mpgedit-$MAJ-${RELEASE}.i386.rpm"
XMPGEDITRPM="xmpgedit-$MAJ-${RELEASE}.i386.rpm"

# Only need mpgedit source rpm, since it is the parent
# of all of the other packages.
#
MPGEDITSRCRPM="mpgedit-$MAJ-${RELEASE}.src.rpm"
XMPGEDITSRCRPM="xmpgedit-$MAJ-${RELEASE}.src.rpm"
MPGEDITMANSRCRPM="mpgedit-man-$MAJ-${RELEASE}.src.rpm"

#if [ ! -f "$RPMBASE/SRPMS/$MPGEDITSRCRPM" ]; then
#echo debug 1 src bad
#echo "$RPMBASE/SRPMS/$MPGEDITSRCRPM"
#fi
#if [ ! -f "$RPMBASE/SRPMS/$MPGEDITMANSRCRPM" ]; then
#echo debug 1 src bad
#echo "$RPMBASE/SRPMS/$MPGEDITMANSRCRPM"
#fi
#if [ ! -f "$RPMBASE/SRPMS/$XMPGEDITSRCRPM" ]; then
#echo debug 1 src bad
#echo "$RPMBASE/SRPMS/$XMPGEDITSRCRPM"
#fi
#exit 1

#if [ ! -f "$RPMBASE/RPMS/i386/$MPGEDITMANRPM" ]; then
#echo debug 1 bad
#  echo "$RPMBASE/RPMS/i386/$MPGEDITMANRPM"
#fi
#if [ ! -f "$RPMBASE/RPMS/i386/$MPGEDITRPM" ]; then
#  echo "$RPMBASE/RPMS/i386/$MPGEDITRPM"
#echo debug 2 bad
#fi
#if [ ! -f "$RPMBASE/RPMS/i386/$XMPGEDITRPM" ]; then
#  echo "$RPMBASE/RPMS/i386/$XMPGEDITRPM"
#echo debug 3 bad
#fi
#exit 1

#Example rpms in build directory...
#
#/usr/src/redhat/RPMS/i386/mpgedit-0.72-0.i386.rpm
#/usr/src/redhat/RPMS/i386/mpgedit-man-0.72-0.i386.rpm
#/usr/src/redhat/RPMS/i386/xmpgedit-0.72-0.i386.rpm
#/usr/src/redhat/SRPMS/mpgedit-0.72-0.src.rpm
#/usr/src/redhat/SRPMS/mpgedit-man-0.72-0.src.rpm
#/usr/src/redhat/SRPMS/xmpgedit-0.72-0.src.rpm

MPGEDITMANTGTRPM="mpgedit-man-$MAJ.${RELEASE}.${TARGET}-${REV}.noarch.rpm"
MPGEDITTGTRPM="mpgedit-$MAJ.${RELEASE}.${TARGET}-${REV}.i386.rpm"
XMPGEDITTGTRPM="xmpgedit-$MAJ.${RELEASE}.${TARGET}-${REV}.i386.rpm"

MPGEDITSRCTGTRPM="mpgedit-$MAJ.${RELEASE}-${REV}.src.rpm"
XMPGEDITSRCTGTRPM="xmpgedit-$MAJ.${RELEASE}-${REV}.src.rpm"
MPGEDITMANSRCTGTRPM="mpgedit-man-$MAJ.${RELEASE}-${REV}.src.rpm"

if [ -f "$RPMBASE/BUILD/${SRC}.tar" ]; then
    rm -f "$RPMBASE/BUILD/${SRC}.tar" 
fi

if [ -d "$RPMBASE/BUILD/$SRC" ]; then
   rm -rf "$RPMBASE/BUILD/$SRC" 
fi

if [ ! -f "${SRC}.exe" ]; then
  echo "ERROR: Unable to find mpgedit source package"
  exit 1
fi
cp "${SRC}.exe" "$RPMBASE/SOURCES"
if [ ! -f $MPGEDIT ]; then
  DIR=`dirname $0`
  if [ -f "$DIR/$MPGEDIT" ]; then
    (
    cd $DIR
    $RPM -ba $MPGEDIT
    $RPM -ba $MPGEDITMAN
    $RPM -ba $XMPGEDIT
    )
  else
    echo "ERROR: Unable to find RPM spec files"
    exit 1
  fi
else
  $RPM -ba $MPGEDIT
  $RPM -ba $MPGEDITMAN
  $RPM -ba $XMPGEDIT
fi

# Copy down the built RPMS into the local build area.
#
if [ -f "$RPMBASE/RPMS/i386/$MPGEDITMANRPM" ]; then
  if [ ! -f "$MPGEDITMANTGTRPM" ]; then
    cp "$RPMBASE/RPMS/i386/$MPGEDITMANRPM" "$MPGEDITMANTGTRPM"
  fi
else
  echo "ERROR: mpgedit man page RPM does not exist"
fi

if [ -f "$RPMBASE/RPMS/i386/$MPGEDITRPM" ]; then
  if [ ! -f "$MPGEDITTGTRPM" ]; then
    cp "$RPMBASE/RPMS/i386/$MPGEDITRPM" "$MPGEDITTGTRPM"
  fi
else
  echo "ERROR: mpgedit RPM does not exist"
fi

if [ -f "$RPMBASE/RPMS/i386/$XMPGEDITRPM" ]; then
  if [ ! -f "$XMPGEDITTGTRPM" ]; then
    cp "$RPMBASE/RPMS/i386/$XMPGEDITRPM" "$XMPGEDITTGTRPM"
  fi
else
  echo "ERROR: xmpgedit RPM does not exist"
fi

# Copy down the mpgedit SRPM into the local build area.
#
if [ -f "$RPMBASE/SRPMS/$MPGEDITSRCRPM" ]; then
  if [ ! -f "$MPGEDITSRCTGTRPM" ]; then
    cp "$RPMBASE/SRPMS/$MPGEDITSRCRPM" "$MPGEDITSRCTGTRPM"
  fi
else
  echo "ERROR: mpgedit SRPM does not exist"
fi

# Copy down the xmpgedit SRPM into the local build area.
#
if [ -f "$RPMBASE/SRPMS/$XMPGEDITSRCRPM" ]; then
  if [ ! -f "$XMPGEDITSRCTGTRPM" ]; then
    cp "$RPMBASE/SRPMS/$XMPGEDITSRCRPM" "$XMPGEDITSRCTGTRPM"
  fi
else
  echo "ERROR: xmpgedit SRPM does not exist"
fi

# Copy down the mpgedit man SRPM into the local build area.
#
if [ -f "$RPMBASE/SRPMS/$MPGEDITMANSRCRPM" ]; then
  if [ ! -f "$MPGEDITMANSRCTGTRPM" ]; then
    cp "$RPMBASE/SRPMS/$MPGEDITMANSRCRPM" "$MPGEDITMANSRCTGTRPM"
  fi
else
  echo "ERROR: mpgedit SRPM does not exist"
fi

exit 0
