#!/bin/sh
#
# Script to create a Mac OS X disk image for xmpgedit
#
if [ `id -u -n` != "admin" ]; then
  echo "This script must be run as admin"
  exit 1
fi

cd gui
rm -f /tmp/xmpgedit_work.dmg /tmp/xmpgedit.dmg


#
# Populate the xmpgedit.app/Contents/Resources directory
#
xpm_files="\
blankdigit_led.xpm      colon_led.xpm           five_led.xpm
one_led.xpm             play.xpm                seven_led.xpm
two_led.xpm             blankpunct_led.xpm      colon_minute_led.xpm
four_led.xpm            pause.xpm               record.xpm
six_led.xpm             volume1.xpm             close.xpm
eight_led.xpm           next_t.xpm              period_led.xpm
record_green.xpm        stop.xpm                zero_led.xpm
colon_hour_led.xpm      eject.xpm               nine_led.xpm
period_second_led.xpm   record_red.xpm          three_led.xpm"

dylib_files="\
../libdecoder_mad.dylib         ../libmpgedit.dylib
../libdecoder_mpg123.dylib      ../libmpgedit_decoder.dylib
../libdecoder_popen.dylib"

macos_files="\
xmpgedit.app/Contents/MacOS/xmpgedit.c
xmpgedit.app/Contents/MacOS/xmpgedit.sh"

info_plist="xmpgedit.app/Contents/Info.plist"

mkdir -p /tmp/xmpgedit.app/Contents/Resources
mkdir -p /tmp/xmpgedit.app/Contents/MacOS
cp -f $xpm_files   /tmp/xmpgedit.app/Contents/Resources
cp -f $dylib_files /tmp/xmpgedit.app/Contents/Resources
cp -f $macos_files /tmp/xmpgedit.app/Contents/MacOS
cp -f $info_plist  /tmp/xmpgedit.app/Contents
cp -f xmpgedit     /tmp/xmpgedit.app/Contents/Resources

(cd /tmp/xmpgedit.app/Contents/MacOS
 make xmpgedit
)

hdiutil create -megabytes 20 /tmp/xmpgedit_work.dmg -layout NONE
vol=`hdid -nomount /tmp/xmpgedit_work.dmg`
echo "debug vol=$vol"

sudo newfs_hfs -v xmpgeditinstall $vol

echo "ejecting disk $vol"
hdiutil eject $vol

echo "mounting disk '/tmp/xmpgedit_work.dmg'"
tmp=`hdid /tmp/xmpgedit_work.dmg`
target=`echo $tmp | awk '{print $2}'`
echo debug "mounted as '$target'"

echo "Copying xmpgedit.app to '$target'"
cp -pr /tmp/xmpgedit.app $target

hdiutil eject $vol

hdiutil convert -format UDZO /tmp/xmpgedit_work.dmg -o /tmp/xmpgedit.dmg

rm -f xmpgedit.dmg
cp /tmp/xmpgedit.dmg .
(cd /tmp && rm -rf /tmp/xmpgedit.app)
ls -l xmpgedit.dmg
